#ifndef TERM_H
#define TERM_H

#include <hush/export.h>
#include <hush/handler.h>

class formatter;
class light_garbage;
class dvstring;

class parser;

class HUSH_API term { // : public handler {
public:
	int n;   // id v token num
	char* s; // functor | var | con v token string -- no collect!
	term* v; // body v token value
public:

term();

term(int tok, const char* val, term* x = 0); // as a token

term(const char* f); // con or var

term(int v); 
term(float f); 
term(double d); 
term(void* m); /// wildcard + assignment 

term(const char* f, const term* t1, const term* t2 = 0);
term(const char* f, const term& t1);
term(const char* f, const term& t1, const term& t2);

term(const term* t, const term* t1, const term* t2 = 0);
virtual ~term();

term(const term*);
term(const term&);

public: //@ virtual constructors

virtual term* mk(const char* f);                // these need to be redefined 
virtual term* mkvar(const char* f); // variable
virtual term* mk(int v); 
virtual term* mk(float f); 
virtual term* mk(double v); 
virtual term* mk(void* m); /// wildcard + assignment 
virtual term* mk(const char* f, const term* t1, const term* t2 = 0);
virtual term* mk(const term* t, const term* t1, const term* t2 = 0); // higher order

public: //@ assignment

term& operator=(const term& t);    // don t bother
term& operator=(const term* t);

public: //@ see <hush/ml/format.h>  -- the default formatter

virtual char* str(formatter* f = 0) const ;

public: //@ for the logic engines

virtual char* operator()(); // a term as an iterator
virtual char* next();

public: //@ output polymorphic operators

operator int() { return n; }
operator char*() { return s; }

public: //@ arity is slightly more complex than you'd expect

virtual  int nargs() const;  // 0,1,2 -- for (binary) formatting
virtual  int arity() const;  // real arity count

virtual  int narity() const;  // auxiliary, for arglists

public: //@ value types

virtual int ival() const;         // undefined in base class
virtual float fval() const;
virtual double dval() const;

virtual char* aval() const;       // default  str()

public: //@ General

virtual char* category() const;   // undefined for now
virtual int level() const;

virtual int type() const;         // must be redefined

virtual term* arg(int i = 0) const; // also subclass specific
virtual term* nth(int i = 0) const; // in arg list

public: //@ meta

virtual int _iswild(const char* s = 0) const;  // define in subclass

virtual int _isnil(const char* s = 0) const;   // define also
virtual int _isnum(const char* s = 0) const;
virtual int _iscon(const char* s = 0) const;
virtual int _isvar(const char* s = 0) const;
virtual int _isfun(const char* s = 0) const;

virtual int _ismeta(const char* s = 0, int nar = -1) const; // 0 1 2

int _iscomma() const { return _isfun(","); }

public: //@ experimental

virtual int has(const char* s = 0) const;       // may be defined in subclass
virtual int has(const term* t) const;

virtual int matches(const char* s) const;
virtual int matches(const term* t) const;


virtual term* varlist() const ;
//virtual char* varname();

virtual term* rename(int);
//virtual void substitute(int,term*);

//virtual char* id() const ;


public: //@ operators
int _isop(const char c) const;
int _opin(const char* s) const;


public: //@ access

virtual char* fun() const;
virtual term* a1() const;
virtual term* a2() const;

//// Lower

virtual term* _body() const;


protected: //@ for allocating functor names

char* _news(const char* f);

public: //@ likely to become obsolete

static parser* _parser;
static parser* _scanner;

public: //@ garbage should not be public

static handler* _garbage;
static void _cleanup();
static void _hold(const term* h); 

public: //@ defaults

static term* _default;
static int _noformat;
static formatter* _formatter;

static int created;
static int deleted;

static void* table;
static void declare(const char*, void*, int flag = 0);
static term* lookup(const char*);
static void destroy();

public: //@ for managing garbage
light_garbage* _light;

virtual void _register( const handler* h) const; // of course, this is not true
virtual void _register( const term* t) const;
virtual void _register( const dvstring* s) const;
virtual void _register( const char* a) const;
};


//@ iostream

//AE?

#include <iostream.h>
#include <hush/string.h>
inline ostream& operator<<(ostream& os, const term& t) {
	char* s = t.str();
	os << (char*) s << endl;
	return os;
	}

//@ convenience functions

inline term* mkterm(int n, const char* s, term* v) { 
	term* x = new term(n,s,v);
	term::_hold(x); return x;
	}
inline term* token(int tok, const char* s, term* x) {
	return mkterm(tok,s,x);
	}

//@ now, the regular ones

inline term* mkterm(const char* s) { 
	term* x = new term(s);
	term::_hold(x); return x;
	}
inline term* mkterm(const char* f, const term* t1, const term* t2 = 0) {
	term* x = new term(f,t1,t2); 
	term::_hold(x); return x;
	}
inline term* mkterm(const char* f, const term& t1) { 
	term* x = new term(f,&t1); 
	term::_hold(x); return x;
	}
inline term* mkterm(const char* f, const term& t1, const term& t2) { 
	term* x = new term(f,&t1,&t2); 
	term::_hold(x); return x;
	}

inline term* mkcon(const char* s) {
	term* x = new term(s);
	term::_hold(x); return x;
	}
inline term* mkvar(const char* s) {
	term* x = new term(s);
	term::_hold(x); return x;
	}
inline term* mkfun(const char* f, const term* t1, const term* t2 = 0) { 
	term* x = new term(f,t1,t2); 
	term::_hold(x); return x;
	}
inline term* mkfun(const char* f, const term& t1) { 
	term* x = new term(f,&t1); 
	term::_hold(x); return x;
	}
inline term* mkfun(const char* f, const term& t1, const term& t2) { 
	term* x = new term(f,&t1,&t2); 
	term::_hold(x); return x;
	}

inline term* mkc(const term* t1, const term* t2) { 
	return mkfun(",",t1,t2);
	}

inline term* mknil() { return mkcon("[]"); }
inline term* mklist(const term* t1, const term* t2) { 
	if (t1 && t2)
	        return mkfun(".",t1,t2);
	else if (t1) return (term*) t1;
	else if (t2) return (term*) t2;
	else return 0;
	}
inline term* mklist(const term& t1, const term& t2) { 
	return mklist(&t1,&t2);
	}

//@ query operators and access

inline int type(const term& t) { return t.type(); }
inline int type(const term* t) { return t?t->type():0; }

inline char* strof(const term& t) { return (char*) strof(&t); }
inline char* strof(const term* t) { 
	if (!t) return 0;
	else if (t->v) return t->v->aval();
	else return t->s;
	}

inline char* funof(const term& t) { return t.fun(); }
inline char* funof(const term* t) { return t?t->fun():0; }

inline term* arg1of(const term& t) { return t.arg(1); }
inline term* arg1of(const term* t) { return t?t->arg(1):0; }

inline term* a1of(const term& t) { return t.arg(1); }
inline term* a1of(const term* t) { return t?t->arg(1):0; }

inline term* a2of(const term& t) { return t.arg(2); }
inline term* a2of(const term* t) { return t?t->arg(2):0; }

inline term* arg2of(const term& t) { return t.arg(2); }
inline term* arg2of(const term* t) { return t?t->arg(2):0; }

//@ tests

inline int iscon(const term& t) { return t._iscon(); }
inline int isnum(const term& t) { return t._isnum(); }
inline int isvar(const term& t) { return t._isvar(); }
inline int isfun(const term& t) { return t._isfun(); }

inline int iscon(const term* t) { return t?t->_iscon():0; }
inline int isnum(const term* t) { return t?t->_isnum():0; }
inline int isvar(const term* t) { return t?t->_isvar():0; }
inline int isfun(const term* t) { return t?t->_isfun():0; }


inline int isnull(const term&) { return 0; } 
inline int isnull(const term* t) { return t != 0; } 

inline int isc(const term& t) { return t._isfun(","); } 
inline int isc(const term* t) { return t?t->_isfun(","):0; } 

inline int iscomma(const term& t) { return t._isfun(","); } 
inline int iscomma(const term* t) { return t?t->_isfun(","):0; } 

inline int isnil(const term& t) { return t._iscon("[]"); } 
inline int isnil(const term* t) { return t?t->_iscon("[]"):0; } 

inline int islist(const term& t) { return t._isfun("."); } 
inline int islist(const term* t) { return t?t->_isfun("."):0; } 

inline int iszero(const term& t) { return t._iscon("0"); } 
inline int iszero(const term* t) { return t?t->_iscon("0"):0; } 

//</>
#endif
